<?php

/**
 * @file
 * Theme for Gantt views.
 */

use Drupal\Core\Url;

/**
 * Template preprocess views gantt.
 *
 * @param array $variables
 *   Array variable.
 */
function template_preprocess_views_view_gantt(array &$variables) {
  $view = $variables['view'];
  $style = $view->style_plugin;
  $setting = $style->options;
  $fieldDate = $view->getHandlers('field')[$setting["start_date"]];
  $dateType = \Drupal::entityTypeManager()
    ->getStorage('date_format')
    ->load($fieldDate["settings"]["format_type"]);
  $patterns = explode(' ', $dateType->getPattern());
  $patterns = array_filter(
    $patterns,
    function ($value) {
      return strlen($value) >= 4;
    }
  );
  $setting['format_date'] = current($patterns);
  $variables["view_id"] = $view->storage->id();
  $variables["display_id"] = $view->current_display;
  if ($setting["cdn"]) {
    $variables['view']->element['#attached']['library'][] = 'gantt/gantt.cdn';
  }
  else {
    $variables['view']->element['#attached']['library'][] = 'gantt/gantt';
  }
  $id = implode('-', ["gantt", $view->id(), $view->current_display]);
  $variables["id_gantt"] = $id;
  $variables['view']->element['#attached']['drupalSettings']['gantt'][$id] = [
    'date_format' => str_replace(
      ['Y', 'm', 'd', 'H', 'i', 's'],
      ['%Y', '%m', '%d', '%H', '%i', '%s'],
      $setting['format_date']
    ),
    'id' => $variables["id_gantt"],
    'owner' => $setting['owner'],
    'ajax' => Url::fromRoute('gantt.ajax', [
      'view_id' => $variables["view_id"],
      'display_id' => $variables["display_id"],
    ])->toString(),
    'import' => Url::fromRoute('gantt.import', [
      'view_id' => $variables["view_id"],
      'display_id' => $variables["display_id"],
    ])->toString(),
  ];
  // Dialog width.
  $variables["data"] = convert_gantt_data($variables['rows'], $setting);
  $variables["owner"] = [];
  $variables["add"] = '';
  // Set session variable for import paragraphs.
  if (!empty($view->argument)) {
    foreach ($view->argument as $key => $arg) {
      if ($key == 'parent_id') {
        $variables["paragraph_parent_id"] = $view->args[$arg->position];
        if (!empty($variables['rows'])) {
          $paragraphFirst = current($variables['rows'])->_entity;
          $variables["paragraph_parent_type"] = current($paragraphFirst->get('parent_type'))[0]->value;
          $variables["paragraph_parent_field_name"] = current($paragraphFirst->get('parent_field_name'))[0]->value;
        }
      }
    }
  }
}

/**
 * {@inheritDoc}
 */
function convert_gantt_data($items, array $setting) {
  $links = [];
  $outData = [];
  $mappingGantt = [
    'text',
    'type',
    'start_date',
    'duration',
    'progress',
    'open',
    'parent',
    'owner',
  ];
  $fieldDateType = 'date';

  foreach ($items as $delta => $item) {
    $entity = $item->_entity;
    $fieldDefinitions = $entity->getFieldDefinitions();
    if (!empty($fieldDefinitions[$setting['start_date']])) {
      $dateSettings = $fieldDefinitions[$setting['start_date']]->getSettings();
      $fieldDateType = $fieldDefinitions[$setting['start_date']]->getType();
    }
    $data[$delta] = ['id' => $entity->id()];
    foreach ($mappingGantt as $ganttField) {
      if (!empty($field_name = $setting[$ganttField])) {
        $value = $entity->get($field_name)->getString();
        if (!empty($value)) {
          $data[$delta][$ganttField] = $value;
        }
        if ($ganttField == 'start_date') {
          if (!empty($entity->get($field_name)->date)) {
            $data[$delta][$ganttField] = $entity->get($field_name)->date->format($setting['format_date']);
          }
          if ($fieldDateType == 'daterange' && $dateValue = $entity->get($field_name)->value) {
            // $endValue = $entity->get($field_name)->end_value;
            $data[$delta][$ganttField] = date($setting['format_date'], strtotime($dateValue));
          }
        }

        // Process owner.
        if ($ganttField == 'owner' && !$entity->get($field_name)->isEmpty()) {
          $values = $entity->get($field_name)->getValue();
          $data[$delta][$ganttField] = [];
          foreach ($values as $value) {
            if (!empty($value->entity)) {
              $data[$delta][$ganttField][] = $value->entity->id();
            }
          }
          // @todo remove later for multiuser.
          $data[$delta][$ganttField] = $value->entity->id();
        }
      }
    }
    // Process Links.
    if (!empty($setting['links'])) {
      $valueLinks = $entity->get($setting['links'])->getValue();
      foreach ($valueLinks as $valueLink) {
        $links[] = [
          'id' => implode('-', [$entity->id(), $valueLink['first']]),
          'source' => $entity->id(),
          'target' => (int) $valueLink['first'],
          'type' => (int) $valueLink['second'],
        ];
      }
    }

  }
  if (!empty($data)) {
    $outData['data'] = $data;
    if (!empty($links)) {
      $outData['links'] = $links;
    }
  }
  return $outData;
}
