/**
 * @file
 */

(($, Drupal, cookies) => {
  Drupal.behaviors.gantt = {
    attach: (context, settings) => {
      $('.showGantt', context).once('gantt').each(function () {
        if (settings.gantt) {
          // Marker today.
          function markerToday() {
            let dateToStr = gantt.date.date_to_str(gantt.config.task_date);

            let id = gantt.addMarker({
              start_date: new Date(),
              css: "today",
              title: dateToStr(new Date())
            });
            setInterval(function () {
              let today = gantt.getMarker(id);
              if (today) {
                today.start_date = new Date();
                today.title = dateToStr(today.start_date);
                gantt.updateMarker(id);
              }
            }, 1000 * 60);
            //Make resize marker for two columns
            gantt.attachEvent("onColumnResizeStart", function (ind, column) {
              if (!column.tree || ind == 0) {
                return;
              }

              setTimeout(function () {
                let marker = document.querySelector(".gantt_grid_resize_area");
                if (!marker) {
                  return;
                }
                let cols = gantt.getGridColumns();
                let delta = cols[ind - 1].width || 0;
                if (!delta) {
                  return;
                }

                marker.style.boxSizing = "content-box";
                marker.style.marginLeft = -delta + "px";
                marker.style.paddingRight = delta + "px";
              }, 1);
            });
          }

          // Undo Redo.
          function undo() {
            gantt.templates.scale_cell_class = function (date) {
              if (!gantt.isWorkTime(date)) {
                return "weekend";
              }
            };
            gantt.templates.timeline_cell_class = function (item, date) {
              if (!gantt.isWorkTime(date)) {
                return "weekend";
              }
            };
            gantt.attachEvent("onBeforeAutoSchedule", function () {
              gantt.message("Recalculating project schedule...");
              return true;
            });
            gantt.attachEvent("onAfterTaskAutoSchedule", function (task, new_date, constraint, predecessor) {
              gantt.message({
                text: "<b>" + task.text + "</b> has been rescheduled to " + gantt.templates.task_date(new_date) + " due to <b>" + predecessor.text + "</b> constraint",
                expire: 4000
              });
            });
          }

          // Expand at bottom.
          function expand() {
            gantt.attachEvent("onCollapse", function () {
              var icon = gantt.toggleIcon;
              if (icon) {
                icon.className = icon.className.replace("fa-compress", "fa-expand");
              }
            });
            gantt.attachEvent("onExpand", function () {
              var icon = gantt.toggleIcon;
              if (icon) {
                icon.className = icon.className.replace("fa-expand", "fa-compress");
              }
            });
            gantt.attachEvent("onTemplatesReady", function () {
              var toggle = document.createElement("i");
              toggle.className = "fa fa-expand gantt-fullscreen";
              gantt.toggleIcon = toggle;
              gantt.$container.appendChild(toggle);
              toggle.onclick = function () {
                gantt.ext.fullscreen.toggle();
              };
            });
          }

          // User owner.
          function byId(list, id) {
            for (let i = 0; i < list.length; i++) {
              if (list[i].key == id) {
                return list[i].label || "";
              }
            }
            return "";
          }

          function owners(owner) {
            gantt.locale.labels.column_owner =
              gantt.locale.labels.section_owner = Drupal.t("Owner");
            gantt.config.resource_property = "staff";
            gantt.serverList("staff", owner);

            gantt.attachEvent("onTaskCreated", function (task) {
              task.staff = [];
              task.calendar_id = null;
              return true;
            });
            gantt.config.columns.unshift({
              name: "owner", width: 80, align: "center", template: function (task) {
                return byId(gantt.serverList('staff'), task.owner)
              }
            });

            gantt.config.lightbox.sections.push(
              {name: "owner", height: 22, map_to: "owner", type: "select", options: gantt.serverList("staff")}
            );
            gantt.templates.rightside_text = function (start, end, task) {
              return byId(gantt.serverList('staff'), task.owner);
            };

            gantt.templates.grid_row_class =
              gantt.templates.task_row_class =
                gantt.templates.task_class = function (start, end, task) {
                  let css = [];
                  if (task.$virtual || task.type == gantt.config.types.project) {
                    css.push("summary-bar");
                  }

                  if (task.owner) {
                    css.push("gantt_resource_task gantt_resource_" + task.owner);
                  }

                  return css.join(" ");
                };

            gantt.attachEvent("onDataRender", function () {
              let styleId = "dynamicGanttStyles";
              let element = document.getElementById(styleId);
              if (!element) {
                element = document.createElement("style");
                element.id = styleId;
                document.querySelector("head").appendChild(element);
              }
              let html = [];
              let resources = gantt.serverList("staff");

              resources.forEach(function (r) {
                html.push(".gantt_task_line.gantt_resource_" + r.key + "{" +
                  "background-color:" + r.backgroundColor + "; " + "color:" + r.textColor + ";" + "}");
                html.push(".gantt_row.gantt_resource_" + r.key + " .gantt_cell:nth-child(1) .gantt_tree_content{" +
                  "background-color:" + r.backgroundColor + "; " + "color:" + r.textColor + ";" + "}");
              });
              element.innerHTML = html.join("");
            });

          }

          // Tooltips.
          function setTooltip() {
            let formatter = gantt.ext.formatters.durationFormatter({
              enter: "day",
              store: "hour",
              format: "day",
              hoursPerDay: 24,
              hoursPerWeek: 40,
              daysPerMonth: 30,
              short: true
            });
            let linksFormatter = gantt.ext.formatters.linkFormatter({durationFormatter: formatter});

            gantt.config.lightbox.sections = [
              {name: "description", height: 70, map_to: "text", type: "textarea", focus: true},
              {
                name: "type", type: "typeselect", map_to: "type", filter: function (name, value) {
                  return !!(value != gantt.config.types.project);
                }
              },
              {name: "time", type: "duration", map_to: "auto", formatter: formatter}
            ];
            gantt.config.lightbox.project_sections = [
              {name: "description", height: 70, map_to: "text", type: "textarea", focus: true},

              {name: "time", type: "duration", readonly: true, map_to: "auto", formatter: formatter}
            ];
            gantt.config.lightbox.milestone_sections = [
              {name: "description", height: 70, map_to: "text", type: "textarea", focus: true},
              {
                name: "type", type: "typeselect", map_to: "type", filter: function (name, value) {
                  return !!(value != gantt.config.types.project);
                }
              },
              {name: "time", type: "duration", single_date: true, map_to: "auto", formatter: formatter}
            ];

            gantt.templates.tooltip_text = function (start, end, task) {
              let html = "<b>" + Drupal.t('Task') + ":</b> " + task.text + "<br/><b>" + Drupal.t("Start") +
                ":</b> " + gantt.templates.tooltip_date_format(start) +
                "<br/><b>" + Drupal.t("End") + ":</b> " + gantt.templates.tooltip_date_format(end) +
                "<br><b>" + Drupal.t("Duration") + ":</b> " + formatter.format(task.duration);
              return html;
            };
          }

          function upload(file, callback, urlImport) {
            gantt.importFromMSProject({
              data: file,
              callback: function (project) {
                if (project) {
                  gantt.clearAll();
                  if (project.config.duration_unit) {
                    gantt.config.duration_unit = project.config.duration_unit;
                  }
                  // mark tasks import.
                  for (let i = 0; i < project.data.data.length; i++) {
                    project.data.data[i].isImport = true;
                  }
                  gantt.parse(project.data);
                  $.post(urlImport, project.data).done(function (data) {
                    console.log("Data imported: ", data);
                  });
                  if (callback) {
                    callback(project);
                  }
                }
              }
            });
          }

          function importMMP(ganttId, urlImport) {
            // Import microsoft project.
            let fileDnD = fileDragAndDrop();
            fileDnD.init(gantt.$container);

            function sendFile(file) {
              fileDnD.showUpload();
              upload(file, function () {
                fileDnD.hideOverlay();
              }, urlImport)
            }

            fileDnD.onDrop(sendFile);
            // Manual Upload file MPP.
            $("#mspImportBtn" + ganttId).click(function (event) {
              event.preventDefault();
              let fileInput = document.getElementById("mspFile" + ganttId);
              if (fileInput.files[0]) {
                sendFile(fileInput.files[0]);
              }
            });
          }
          function childTogetherParent() {
            gantt.attachEvent("onTaskDrag", function (id, mode, task, original) {
              var modes = gantt.config.drag_mode;
              if (mode == modes.move) {
                var diff = task.start_date - original.start_date;
                gantt.eachTask(function (child) {
                  child.start_date = new Date(+child.start_date + diff);
                  child.end_date = new Date(+child.end_date + diff);
                  gantt.refreshTask(child.id, true);
                }, id);
              }
            });
//rounds positions of the child items to scale
            gantt.attachEvent("onAfterTaskDrag", function (id, mode, e) {
              var modes = gantt.config.drag_mode;
              if (mode == modes.move) {
                var state = gantt.getState();
                gantt.eachTask(function (child) {
                  child.start_date = gantt.roundDate({
                    date: child.start_date,
                    unit: state.scale_unit,
                    step: state.scale_step
                  });
                  child.end_date = gantt.calculateEndDate(child.start_date,
                    child.duration, gantt.config.duration_unit);
                  gantt.updateTask(child.id);
                }, id);
              }
            });
          }
          $(document).ready(function () {
            if (typeof gantt == 'undefined') {
              console('Cant load Gantt');
              return false;
            }
            var paragraphGantt = [];
            var isFirst = true;

            gantt.plugins({
              fullscreen: true,
              marker: true,
              critical_path: true,
              auto_scheduling: true,
              tooltip: true,
              undo: true
            });
            childTogetherParent();
            markerToday();
            expand();
            setTooltip();
            undo();
            gantt.config.auto_types = true;
            gantt.config.duration_unit = "day";
            gantt.config.order_branch = "marker";
            gantt.config.order_branch_free = true;
            gantt.config.grid_resize = true;

            gantt.config.work_time = true;
            gantt.config.auto_scheduling = true;
            gantt.config.auto_scheduling_strict = true;
            gantt.config.auto_scheduling_compatibility = true;

            gantt.config.fit_tasks = true;
            gantt.ext.zoom.init(zoomConfig);
            gantt.ext.zoom.setLevel("day");
            gantt.$zoomToFit = false;

            $.each(settings.gantt, function (key, ganttField) {
              gantt.config.date_format = ganttField.date_format;
              gantt.config.date_grid = ganttField.date_format;
              let configAjax = {
                url: ganttField.ajax,
                mode: "POST"
              };
              if ($('#' + ganttField.id).length) {
                let dataParagraph = $('#' + ganttField.id).data();
                configAjax.url += '?' + $.param( dataParagraph );
                ganttField.import += '?' + $.param( dataParagraph );
              }
              if (isFirst) {
                if ($('#' + ganttField.id).length) {
                  gantt.init(ganttField.id);
                }
                if (dataGantt[ganttField.id] && dataGantt[ganttField.id].data !== undefined) {
                  gantt.parse(dataGantt[ganttField.id]);
                }
                if (ganttField.owner && typeof ownerGantt != 'undefined') {
                  owners(ownerGantt[ganttField.id]);
                }
                let dp = gantt.createDataProcessor(configAjax);
                importMMP(ganttField.id, ganttField.import);
                isFirst = false;
              } else {
                // Multiple Charts on a Page. Unsupported cdn version.
                // it's just show chart, @todo later.
                Gantt[key] = Gantt.getGanttInstance();
                //Gantt[key].config.date_grid = settings.date_format;
                Gantt[key].init(ganttField.id);
                if (Gantt[ganttField.id].data !== undefined) {
                  Gantt[key].parse(Gantt[ganttField.id]);
                }
                let dp = Gantt[key].createDataProcessor(configAjax);
              }
            });
          });
        }
      });
    },
  };
})(jQuery, Drupal, window.Cookies);
