<?php

namespace Drupal\gantt\Controller;

use Drupal\Core\Controller\ControllerBase;
use Drupal\Core\Datetime\Entity\DateFormat;
use Drupal\paragraphs\Entity\Paragraph;
use Drupal\views\Views;
use Symfony\Component\HttpFoundation\JsonResponse;

/**
 * Returns responses for Gantt routes.
 */
class GanttController extends ControllerBase {


  /**
   * Query POST.
   *
   * @var array
   */
  private $post = [];

  /**
   * Query GET.
   *
   * @var array
   */
  private $get = [];

  /**
   * {@inheritDoc}
   */
  public function __construct() {
    $request = \Drupal::request();
    $this->post = $request->request->all();
    $this->get = $request->query->all();
  }

  /**
   * {@inheritdoc}
   */
  public function loadView($view_id, $display_id, &$fieldSetting = []) {
    $view = Views::getView($view_id);
    $view->setDisplay($display_id);
    $view->initDisplay();
    $view->initHandlers();
    // $baseTable = $view->storage->get('base_table');
    $fieldSetting = $view->getStyle()->options;
    $entity_type = 'node';
    $filter = $view->filter;
    if (!empty($filter['type'])) {
      $entity_type = $filter['type']->getEntityType();
      $fieldSetting['bundle'] = key($filter['type']->value);
    }
    return $this->entityTypeManager()->getStorage($entity_type);
  }

  /**
   * Add new / Edit / Delete Paragraphs.
   */
  public function ajax($view_id, $display_id) {
    $fieldSetting = [];
    $entityType = $this->loadView($view_id, $display_id, $fieldSetting);
    $bundle = !empty($fieldSetting['bundle']) ? $fieldSetting['bundle'] : 'gantt';
    $action = !empty($this->post["!nativeeditor_status"]) ? $this->post["!nativeeditor_status"] : FALSE;
    $output = [];
    if (!$action) {
      return new JsonResponse($output);
    }
    if ($this->get['gantt_mode'] == 'tasks') {
      $output = $this->actionTask($action, $entityType, $bundle, $fieldSetting);
    }
    if ($this->get['gantt_mode'] == 'links' && !empty($fieldSetting["links"])) {
      $output = $this->actionLinks($action, $entityType, $fieldSetting);
    }
    return new JsonResponse($output);
  }

  /**
   * Import file MPP.
   */
  public function import($view_id, $display_id, Request $request = NULL) {
    $fieldSetting = [];
    $entityType = $this->loadView($view_id, $display_id, $fieldSetting);
    $bundle = !empty($fieldSetting['bundle']) ? $fieldSetting['bundle'] : 'gantt';
    $action = 'inserted';
    $output = ['action' => $action];
    $mapping = [];
    if (!empty($this->post["data"])) {
      $rawData = $this->post["data"];
      $rawLinks = $this->post["links"];
      $this->post = [];
      foreach ($rawData as $data) {
        $ganttData[$data['id']] =
        $this->post = [
          'text' => $data['text'],
          'duration' => $data['duration'],
          'progress' => $data['progress'],
          'open' => $data['open'],
          'parent' => $data['parent'],
          'type' => !empty($data['type']) ? $data['type'] : '',
          'start_date' => $data['$raw']["Start"],
          'end_date' => $data['$raw']["Finish"],
        ];
        if (!empty($mapping[$data['parent']])) {
          $this->post['parent'] = $mapping[$data['parent']];
        }
        $output = $this->actionTask($action, $entityType, $bundle, $fieldSetting);
        $mapping[$data['id']] = !empty($output['tid']) ? $output['tid'] : FALSE;
      }

      foreach ($rawLinks as $data) {
        $this->post = [
          'source' => $mapping[$data['source']],
          'target' => $mapping[$data['target']],
          'type' => $data['type'],
        ];
        $output = $this->actionLinks($action, $entityType, $fieldSetting);
      }
    }

    return new JsonResponse($output);
  }

  /**
   * Convert date return with format Y-m-d.
   */
  protected function formatDate($date, $type = 'date', $format = 'Y-m-d H:i:s') {
    $drupalShortFormat = DateFormat::load('short')->getPattern();
    if ($type == 'date') {
      $format = current(explode(' ', $format));
      $drupalShortFormat = current(explode(' ', $drupalShortFormat));
    }
    $convertDateTime = \DateTime::createFromFormat($drupalShortFormat, $date);
    if (!$convertDateTime) {
      $date = str_replace('/', '-', $date);
      $timestamp = strtotime($date);
      return date($format, $timestamp);
    }
    return $convertDateTime->format($format);
  }

  /**
   * Process Task.
   *
   * {@inheritDoc}
   */
  private function actionTask($action, $entityType, $bundle, $field_settings) {
    $output = ['action' => $action];
    if (!empty($this->post['isImport'])) {
      return new JsonResponse($output);
    }
    if ($entityType->getEntityTypeId() == "paragraph") {
      return $this->actionParagraphsTask($action, $entityType, $field_settings);
    }
    else {
      return $this->actionEntityTask($action, $entityType, $bundle, $field_settings);
    }
  }

  /**
   * Process Entity Task.
   *
   * {@inheritDoc}
   */
  private function actionEntityTask($action, $entityType, $bundle, $field_settings) {

    $output['action'] = $action;
    if ($action == 'inserted') {
      $entity = $entityType->create([
        'type' => $bundle,
        'title' => $this->post['text'],
        'uid' => \Drupal::currentUser()->id(),
      ]);
    }
    else {
      $entity = $entityType->load($this->post["id"]);
    }
    if ($action == 'deleted') {
      $entity->delete();
      return $output;
    }
    $entity = $this->setFieldValue($entity, $field_settings);
    switch ($action) {
      case 'inserted':
        $entity->isNew();
        break;
    }
    $entity->save();
    $output['tid'] = $entity->id();
    return $output;
  }

  /**
   * {@inheritDoc}
   */
  private function setFieldValue($entity, $field_settings) {
    $fieldDefinitions = $entity->getFieldDefinitions();
    $fieldType = $dateType = 'date';
    if (is_array($fieldDefinitions) && !empty($fieldDefinitions[$field_settings['start_date']])) {
      $dateSettings = $fieldDefinitions[$field_settings['start_date']]->getSettings();
      $dateType = $dateSettings["datetime_type"];
      $fieldType = $fieldDefinitions[$field_settings['start_date']]->getType();
    }
    $dateFormat = $fieldType == 'daterange' ? 'Y-m-d\TH:i:s' : 'Y-m-d H:i:s';
    $start_date = $this->formatDate($this->post['start_date'], $dateType, $dateFormat);
    $end_date = $this->formatDate($this->post['end_date'], $dateType, $dateFormat);
    if ($fieldType == 'daterange') {
      $start_date = [
        'value' => $start_date,
        'end_value' => $end_date,
      ];
    }
    $entity->set($field_settings['text'], $this->post['text']);
    $entity->set($field_settings['duration'], $this->post['duration']);
    $entity->set($field_settings['start_date'], $start_date);
    if (!empty($this->post['progress'])) {
      $entity->set($field_settings['progress'], $this->post['progress']);
    }
    $parent = !empty($this->post['parent']) ? $this->post['parent'] : 0;
    $entity->set($field_settings['parent'], $parent);
    $open = !empty($this->post['open']) ? $this->post['open'] : 0;
    $entity->set($field_settings['open'], $open);
    if (!empty($field_settings["owner"]) && !empty($this->post['owner'])) {
      $owner = [];
      if (is_array($this->post['owner'])) {
        foreach ($this->post['owner'] as $owner_id) {
          $owner[] = ['target_id' => $owner_id];
        }
      }
      else {
        $owner = [['target_id' => $this->post['owner']]];
      }
      $entity->set($field_settings['owner'], $owner);
    }
    return $entity;
  }

  /**
   * Process Paragraphs Task.
   *
   * {@inheritDoc}
   */
  private function actionParagraphsTask($action, $paragraphsDef, $field_settings) {
    $paragraph = !empty($this->post["id"]) ? Paragraph::load($this->post["id"]) : FALSE;
    $entity_field = 'field_gantt';
    $output = ['action' => $action];
    if (empty($paragraph) && !empty($this->post["parent"])) {
      $paragraph = Paragraph::load($this->post["parent"]);
    }
    if (!empty($paragraph)) {
      $entity = $paragraph->getParentEntity();
      $entity_field = $paragraph->get('parent_field_name')->value;
      $paragraphType = $paragraph->getParagraphType()->id();
    }
    else {
      $paragraphType = $this->getParagaphTypeFromView();
      $paragraph_parent_field_name = \Drupal::request()->query->get('paragraph_parent_field_name');
      if (!empty($paragraph_parent_field_name)) {
        $entity_field = $paragraph_parent_field_name;
      }
      $parent_type = \Drupal::request()->query->get('paragraph_parent_type');
      $parent_id = \Drupal::request()->query->get('paragraph_parent_id');
      if (!empty($parent_type) && !empty($parent_id)) {
        $entity = \Drupal::entityTypeManager()
          ->getStorage($parent_type)
          ->load($parent_id);
      }
    }
    if ($action == 'inserted') {
      $paragraph = Paragraph::create($para = ['type' => $paragraphType]);
    }

    if ($action == 'deleted') {
      $getField = $entity->get($entity_field);
      $array_of_referenced_items = $getField->getValue();
      $index_to_remove = array_search($this->post['id'], array_column($array_of_referenced_items, 'target_id'));
      $getField->removeItem($index_to_remove);
      $entity->save();
      $paragraph->delete();
      return new JsonResponse($output);
    }

    $fieldDefinitions = $paragraph->getFieldDefinitions();
    $fieldType = $dateType = 'date';
    if (!empty($fieldDefinitions[$field_settings['start_date']])) {
      $dateSettings = $fieldDefinitions[$field_settings['start_date']]->getSettings();
      $dateType = $dateSettings["datetime_type"];
      $fieldType = $fieldDefinitions[$field_settings['start_date']]->getType();
    }
    $dateFormat = $fieldType == 'daterange' ? 'Y-m-d\TH:i:s' : 'Y-m-d H:i:s';
    $start_date = $this->formatDate($this->post['start_date'], $dateType, $dateFormat);
    $end_date = $this->formatDate($this->post['end_date'], $dateType, $dateFormat);
    if ($fieldType == 'daterange') {
      $start_date = [
        'value' => $start_date,
        'end_value' => $end_date,
      ];
    }
    $paragraph->set($field_settings['text'], $this->post['text']);
    $paragraph->set($field_settings['duration'], $this->post['duration']);
    $paragraph->set($field_settings['start_date'], $start_date);
    if (!empty($this->post['progress'])) {
      $paragraph->set($field_settings['progress'], $this->post['progress']);
    }
    $parent = !empty($this->post['parent']) ? $this->post['parent'] : 0;
    $paragraph->set($field_settings['parent'], $parent);
    $open = !empty($this->post['open']) ? $this->post['open'] : 0;
    $paragraph->set($field_settings['open'], $open);
    if (!empty($field_settings["owner"]) && !empty($this->post['owner'])) {
      $owner = [];
      if (is_array($this->post['owner'])) {
        foreach ($this->post['owner'] as $owner_id) {
          $owner[] = ['target_id' => $owner_id];
        }
      }
      else {
        $owner = [['target_id' => $this->post['owner']]];
      }
      $paragraph->set($field_settings['owner'], $owner);
    }
    if ($action == 'inserted') {
      $paragraph->isNew();
    }
    $paragraph->save();

    if ($action == 'inserted') {
      $output['tid'] = $paragraph->id();
      $fieldValue = $entity->get($entity_field)->getValue();
      $fieldValue[] = [
        'target_id' => $paragraph->id(),
        'target_revision_id' => $paragraph->getRevisionId(),
      ];
      $entity->set($entity_field, $fieldValue);
      $entity->save();
    }

    return $output;
  }

  /**
   * {@inheritDoc}
   */
  private function getParagaphTypeFromView() {
    $parameters = \Drupal::routeMatch()->getParameters();
    $view_id = $parameters->get("view_id");
    $display_id = $parameters->get("display_id");
    $view = Views::getView($view_id);
    $view->setDisplay($display_id);
    $view->initDisplay();
    $view->initHandlers();
    $filter = $view->filter;
    if (!empty($filter["type"])) {
      return key($filter['type']->value);
    }
    return FALSE;
  }

  /**
   * {@inheritDoc}
   */
  public function actionLinks($action, $entityType, $field_settings) {
    $output = ['action' => $action];
    if (!empty($this->post['isImport'])) {
      return new JsonResponse($output);
    }
    $source = !empty($this->post['source']) ? $this->post['source'] : FALSE;
    $target = !empty($this->post['target']) ? $this->post['target'] : FALSE;
    $type = !empty($this->post['type']) ? $this->post['type'] : FALSE;
    if ($action == 'deleted') {
      [$source, $target] = explode('-', $this->post['id']);
    }
    if (!$source) {
      return FALSE;
    }
    $entity = $entityType->load($source);
    $linkValues = $entity->get($field_settings['links'])->getValue();

    if ($action == 'inserted' || !empty($this->post['isImport'])) {
      $linkValues[] = ['first' => $target, 'second' => $type];
    }
    if ($action == 'deleted') {
      foreach ($linkValues as $delta => $linkValue) {
        if ($linkValue['first'] == $target) {
          unset($linkValues[$delta]);
        }
      }
    }
    $entity->set($field_settings['links'], $linkValues);
    $entity->save();

    return new JsonResponse($output);
  }

}
